<?php

namespace Nwb\UrlToPdf;

use Exception;
use Throwable;

class UrlToPdf
{
    private string $apiUrl = 'https://url-to-pdf.kub.nwb.fr';
    private array $urlsAlreadyTested = [];
    private array $downloadedFiles = [];

    /**
     * @var string|null
     */
    private ?string $accessKey;

    /**
     * UrlToPdf constructor.
     * @param array $options Liste des options de configuration
     * @throws Exception
     */
    public function __construct(array $options = [])
    {
        $key = $options['accessKey'] ?? $_ENV['URL_TO_PDF_ACCESS_KEY'] ?? null;
        if (!$key) {
            throw new Exception('Missing access key');
        }
        $this->accessKey = $key;
        $this->apiUrl = $options['apiUrl'] ?? $_ENV['URL_TO_PDF_APIURL'] ?? $this->apiUrl;
    }

    /**
     * Requête curl vers l'API
     * 
     * @param string $url contient la fin de l'url pour appeler la bonne méthode de l'API (ex : /get-pdf)
     * @param array $options Liste des options de configuration, notamment $options['url'] l'url vers la page html à convertir en PDF
     * @return false|resource contenu du PDF
     * @throws Exception si la page est vide
     * @throws Exception si le code HTTP n'est pas 200
     * @throws Exception si le contenu de la réponse n'est pas un PDF
     * @throws Exception si le contenu de la réponse est vide
     */
    private function request(string $url, array $options)
    {
        if (empty($options['url'])) {
            throw new Exception('Missing url');
        }

        // On construit l'url de l'HTML à transmettre à l'API si authentification basique au server (ex: https://username:password@url)
        if (!empty($options['auth']['username']) && !empty($options['auth']['password'])) {
            $username = $options['auth']['username'];
            $password = $options['auth']['password'];
            $options['url'] = str_replace('://', '://' . $username . ':' . $password . '@', $options['url']);
        }

        $urlMustHave200 = $options['urlMustHave200'] ?? false;

        // --- Toujours vérifier que la page HTML n'est pas vide, comme avant ---
        // On fait un cURL GET complet (pas partiel) pour récupérer le HTML.
        $ch1 = curl_init($options['url']);
        curl_setopt($ch1, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch1, CURLOPT_FOLLOWLOCATION, true);
        $pageContent = curl_exec($ch1);

        if ($pageContent === false) {
            throw new Exception("Erreur cURL lors de la récupération du fichier html : " . curl_error($ch1));
        }

        if (empty($pageContent)) {
            curl_close($ch1);
            throw new Exception("La page html récupérée n'a pas de contenu.");
        }

        $httpCode = curl_getinfo($ch1, CURLINFO_HTTP_CODE);
        $contentType = curl_getinfo($ch1, CURLINFO_CONTENT_TYPE);
        curl_close($ch1);

        if ($urlMustHave200) {
            // Si on a besoin de 200, on vérifie
            if ($httpCode !== 200) {
                throw new Exception('Error: ' . $httpCode);
            }

            if (strpos($contentType, 'html') === false) {
                throw new Exception("Le contenu récupéré n'est pas de type HTML.");
            }
        }

        // Maintenant on appelle l'API
        $ch = curl_init();
        $headers = [
            'Content-Type: application/json',
            'ACCESS-KEY: ' . $this->accessKey,
        ];

        curl_setopt($ch, CURLOPT_URL, $this->apiUrl . $url);
        curl_setopt($ch, CURLOPT_POST, true);
        curl_setopt($ch, CURLOPT_POSTFIELDS, json_encode($options));
        curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        $response = curl_exec($ch);

        if ($response === false) {
            throw new Exception("Erreur cURL : " . curl_error($ch));
        }

        $httpCode = curl_getinfo($ch, CURLINFO_HTTP_CODE);
        $contentType = curl_getinfo($ch, CURLINFO_CONTENT_TYPE);
        curl_close($ch);

        if ($httpCode !== 200) {
            if (strpos($contentType, 'application/json') !== false) {
                $errorText = json_decode($response, true);
                if (!empty($errorText)) {
                    $message = $errorText['message'] ?? ($errorText['message'] ?? print_r($errorText, true));
                    throw new Exception('Erreur de l\'API : ' . $message);
                }
            } else {
                if ($response) {
                    throw new Exception('Erreur cURL : ' . $response);
                }
            }
            throw new Exception('Code d\'erreur http : ' . $httpCode);
        }

        if (empty($response) || substr($response, 0, 4) !== "%PDF") {
            throw new Exception('La réponse retournée est vide ou n\'est pas un PDF.');
        }

        return $response;
    }

    public static function hashUrl(string $url): string
    {
        return strtoupper(hash('md5', $url, false));
    }

    /**
     * Enregistre le contenu dans un fichier sur le disque.
     *
     * @param string $content Contenu à enregistrer.
     * @param string $destination Chemin de destination du fichier.
     * @return bool
     */
    public function saveTo(string $content, string $destination): bool
    {
        return file_put_contents($destination, $content) !== false;
    }

    /**
     * Retourne un PDF à partir d'une URL et des options
     * 
     * @param array $options Liste des options de configuration
     * @return false|resource contenu du PDF
     */
    public function getPdf(?array $options)
    {
        $options = $options ?? [
            'url' => 'https://normandiewebschool.fr/cursus/bachelor-developpement-web/',
            'urlMustHave200' => false,
            'loadingOptions' => ['waitUntil' => 'networkidle2'],
            'pdfOptions' => ['format' => 'a4']
        ];

        return $this->request('/get-pdf', $options);
    }

    public function getPdfFromHtml(
        string $htmlContent,
        string $projectRootPath,
        string $webPath,
        string $webPathUrl,
        array $loadingOptions = [],
        array $pdfOptions = [],
        array $authOptions = [],
        bool $isTestMode = false,
        $domaineException = []
    ): string {
        if (!$htmlContent) {
            throw new Exception("Le contenu HTML est vide.");
        }

        if (!is_dir($webPath)) {
            throw new Exception("Le répertoire web n'existe pas : " . $webPath);
        }

        if (!$webPathUrl || strlen($webPathUrl) < 8) {
            throw new Exception("Le chemin web n'est pas valide : " . $webPathUrl);
        }

        if (substr($webPath, -1) !== '/') {
            $webPath .= '/';
        }

        if (substr($webPathUrl, -1) !== '/') {
            $webPathUrl .= '/';
        }

        $webPathUrl = trim($webPathUrl);

        // Vérification de l'url formée
        if (!$this->isValidUrl($webPathUrl)) {
            throw new Exception("L'url n'est pas valide pour : " . $webPathUrl);
        }

        $webPathFinal = $webPath;

        $name_dir = md5($htmlContent);
        $tempWebPath = $webPath . $name_dir . '/';
        $tempWebPathUrl = $webPathUrl . $name_dir . '/';

        $pdfOptions = $pdfOptions ?? ['format' => 'a4'];
        $pdfOptions['noBreakTableRows'] = $pdfOptions['noBreakTableRows'] ?? true;
        $pdfOptions['urlMustHave200'] = $pdfOptions['urlMustHave200'] ?? false;

        $cleanHtml = $this->getCleanHtml($htmlContent, $projectRootPath, $tempWebPath, $tempWebPathUrl, $pdfOptions, $isTestMode, $domaineException);

        if (!is_dir($tempWebPath)) {
            if (!mkdir($tempWebPath, 0777, true)) {
                throw new Exception("Le répertoire web n'a pas pu être créé.");
            }
        }

        $fileName = 'pdf.html';
        $filePath = $tempWebPath . $fileName;
        file_put_contents($filePath, $cleanHtml);
        chmod($filePath, 0777);

        $loadingOptions = $loadingOptions ?? ['waitUntil' => 'networkidle2', 'timeout' => 30000];

        if (!$isTestMode) {
            $pdf = $this->getPdf([
                'url' => $tempWebPathUrl . $fileName,
                'urlMustHave200' => $pdfOptions['urlMustHave200'],
                'loadingOptions' => $loadingOptions,
                'pdfOptions' => $pdfOptions,
                'auth' => $authOptions
            ]);
        } else {
            file_put_contents($tempWebPath . 'TEST_' . $fileName, $cleanHtml);
            $encodedHtml = base64_encode($cleanHtml);
            $encodedHtml = substr($encodedHtml, 0, 3000);
            $pdf = $this->getPdf([
                'url' => 'https://n8n.kub.nwb.fr/webhook/Base64ToHtml?htmlB64=' . $encodedHtml,
                'pdfOptions' => $pdfOptions,
                // Le test mode ne nécessite pas urlMustHave200
            ]);
        }

        file_put_contents($filePath, $pdf);
        rename($filePath, $tempWebPath . 'pdf.pdf');
        $filePath = $tempWebPath . 'pdf.pdf';

        $hashFileName = md5_file($filePath);
        $filePathReturn = $webPathFinal . $hashFileName . '.pdf';
        copy($filePath, $filePathReturn);

        if (!$isTestMode) {
            $this->deleteTempImages($tempWebPath);
        }

        return $filePathReturn;
    }

    private function extractAndCheckBackgroundUrls(string $html): array
    {
        $regex = '/background(?:-image)?:\s*url\((["\']?)(.*?)\1\)/i';
        $uniqueUrls = [];
        if (preg_match_all($regex, $html, $matches)) {
            foreach ($matches[2] as $url) {
                $url = trim($url);
                if (!in_array($url, $uniqueUrls)) {
                    $uniqueUrls[] = $url;
                }
            }
        }
        return $uniqueUrls;
    }

    /**
     * @param string $url URL à vérifier pour récupérer les informations de l'image.
     * @return array{http_code:int, content_type:string}
     * @throws Exception
     */
    public function getImageInfo(string $url): array
    {
        // Tentative avec HEAD
        $ch = curl_init($url);
        // Paramètres cURL pour HEAD
        curl_setopt($ch, CURLOPT_NOBODY, true);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch, CURLOPT_FOLLOWLOCATION, true);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);
        curl_setopt($ch, CURLOPT_TIMEOUT, 10);
        curl_setopt($ch, CURLOPT_USERAGENT, 'Mozilla/5.0');
        curl_exec($ch);
        $code = curl_getinfo($ch, CURLINFO_HTTP_CODE);
        $type = curl_getinfo($ch, CURLINFO_CONTENT_TYPE);
        curl_close($ch);
        // Si HEAD réussit en 200 avec un type
        if ($code === 200 && $type)
            return ['http_code' => 200, 'content_type' => $type];

        // Tentative GET partielle
        $ch = curl_init($url);
        // Paramètres cURL pour GET partielle (0-1024 octets)
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch, CURLOPT_FOLLOWLOCATION, true);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);
        curl_setopt($ch, CURLOPT_TIMEOUT, 10);
        curl_setopt($ch, CURLOPT_USERAGENT, 'Mozilla/5.0');
        curl_setopt($ch, CURLOPT_HEADER, true);
        curl_setopt($ch, CURLOPT_RANGE, '0-1024');
        curl_exec($ch);
        $code = curl_getinfo($ch, CURLINFO_HTTP_CODE);
        $type = curl_getinfo($ch, CURLINFO_CONTENT_TYPE);
        curl_close($ch);

        // Si GET partielle réussit (206 attendu) on considère l'image accessible
        if ($code === 206 && $type)
            return ['http_code' => 200, 'content_type' => $type];

        throw new Exception("Impossible d'accéder à cette image : " . $url);
    }


    private function getFileName($filePath)
    {
        $path = parse_url($filePath, PHP_URL_PATH);
        return basename($path);
    }

    private function getExtensionFromContentType($contentType, $nameFile = null)
    {
        $mimeTypes = [
            'image/jpeg' => 'jpg',
            'image/png' => 'png',
            'image/gif' => 'gif',
            'image/bmp' => 'bmp',
            'image/webp' => 'webp',
            'image/svg' => 'svg'
        ];

        if (!empty($mimeTypes[$contentType])) {
            return $mimeTypes[$contentType];
        } elseif (!empty($nameFile)) {
            $extension = pathinfo($nameFile, PATHINFO_EXTENSION);
            if ($extension) {
                return strtolower($extension);
            }
        }
        return null;
    }

    /**
     * @param string $url URL à vérifier
     * @return bool true si le fichier est accessible
     */
    public function urlExists(string $url): bool
    {
        if (isset($this->urlsAlreadyTested[$url])) {
            return $this->urlsAlreadyTested[$url];
        }

        // Tentative avec HEAD
        $ch = curl_init($url);
        curl_setopt($ch, CURLOPT_NOBODY, true);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch, CURLOPT_FOLLOWLOCATION, true);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);
        curl_setopt($ch, CURLOPT_TIMEOUT, 10);
        curl_setopt($ch, CURLOPT_USERAGENT, 'Mozilla/5.0');
        curl_exec($ch);
        $code = curl_getinfo($ch, CURLINFO_HTTP_CODE);
        curl_close($ch);

        if ($code === 200) {
            $this->urlsAlreadyTested[$url] = true;
            return true;
        }

        // Tentative GET partielle
        $ch = curl_init($url);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch, CURLOPT_FOLLOWLOCATION, true);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);
        curl_setopt($ch, CURLOPT_TIMEOUT, 10);
        curl_setopt($ch, CURLOPT_USERAGENT, 'Mozilla/5.0');
        curl_setopt($ch, CURLOPT_HEADER, true);
        curl_setopt($ch, CURLOPT_RANGE, '0-1024');
        curl_exec($ch);
        $code = curl_getinfo($ch, CURLINFO_HTTP_CODE);
        curl_close($ch);

        $this->urlsAlreadyTested[$url] = ($code === 206);

        return $this->urlsAlreadyTested[$url];
    }

    private function isDomainAllowed($url, $allowedDomains)
    {
        // Convertir l'URL en minuscules
        $url = strtolower($url);
        // Parcourir les domaines autorisés
        foreach ($allowedDomains as $allowedDomain) {
            // Convertir le domaine autorisé en minuscules
            $allowedDomain = strtolower($allowedDomain);
            // Vérifier si le domaine est présent dans l'URL
            if (strpos($url, $allowedDomain) !== false)
                return true;
        }
        return false;
    }

    /**
     * Cette fonction vérifie la validité d'une URL, y compris les cas sans TLD ou avec localhost.
     * @param string $url L'URL à valider
     * @return bool true si l'URL est valide, false sinon
     */
    public function isValidUrl($url): bool
    {
        // Nouveau pattern plus permissif pour gérer les URL S3 et domaines complexes.
        // On autorise plusieurs sous-domaines, des TLD plus longs ainsi que différents protocoles.
        $pattern = '/^(https?|ftp|s3|file|gopher|ws|wss):\/\/(localhost|[a-z0-9\-]+(\.[a-z0-9\-]+)*)(:[0-9]+)?([\/\w.\-%]*)*(\?[\w.\-%&=]*)?(#[\w\-]*)?$/i';

        // Retour anticipé : si pas de correspondance, retourner false
        if (!preg_match($pattern, $url))
            return false;

        return true;
    }

    /**
     * Télécharge un fichier à partir d'une URL et le sauvegarde à un chemin donné.
     *
     * @param string $url URL du fichier à télécharger.
     * @param string $destination Chemin où sauvegarder le fichier.
     * @throws Exception si le téléchargement échoue.
     */
    private function downloadFile(string $url, string $destination): void
    {
        if (isset($this->downloadedFiles[$destination])) {
            return;
        }

        $ch = curl_init($url);
        $fp = fopen($destination, 'wb');

        if ($fp === false) {
            throw new Exception("Impossible de créer le fichier à l'emplacement : $destination");
        }

        curl_setopt($ch, CURLOPT_FILE, $fp);
        curl_setopt($ch, CURLOPT_FOLLOWLOCATION, true);
        curl_setopt($ch, CURLOPT_TIMEOUT, 30);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);

        $success = curl_exec($ch);

        if ($success === false) {
            $error = curl_error($ch);
            curl_close($ch);
            fclose($fp);
            throw new Exception("Erreur lors du téléchargement du fichier : $error");
        }

        $httpCode = curl_getinfo($ch, CURLINFO_HTTP_CODE);
        curl_close($ch);
        fclose($fp);

        if ($httpCode !== 200) {
            unlink($destination); // Supprime le fichier en cas d'erreur HTTP.
            throw new Exception("Le téléchargement a échoué avec le code HTTP : $httpCode");
        }
        $this->downloadedFiles[$destination] = true;
    }

    /**
     * Retourne le html passé en paramètre avec les path des fichiers accessibles pour l'API Url to PDF
     * (méthode historique : copie des fichiers dans un dossier + balises locales pointant vers ce dossier).
     *
     * @param string $htmlIn
     * @param string $projectRootPath (ex: /var/www/html/)
     * @param string $webPath (ex: /var/www/html/public-data/tmp/url_to_pdf/)
     * @param string $webPathUrl (ex: https://www.nwb.fr/public-data/tmp/url_to_pdf/)
     * @param array  $pdfOptions      Liste des options de configuration dédiées au module et à la conversion html
     * @param bool   $isTestMode      Indique si on est en mode test, dans ce cas on ne n'insère pas les path sous forme d'url
     * @param array  $domaineException Liste des domaines à ne pas traiter
     *
     * @return string
     * @throws Exception si l'image n'existe pas ou si l'URL n'est pas valide
     */
    public function getCleanHtml(
        string $htmlIn,
        string $projectRootPath,
        string $webPath,
        string $webPathUrl,
        array $pdfOptions = [],
        bool $isTestMode = false,
        array $domaineException = []
    ): string {
        // Appel de la méthode commune avec $useBase64 = false
        return $this->doCleanHtml(
            $htmlIn,
            $projectRootPath,
            $webPath,
            $webPathUrl,
            $pdfOptions,
            $isTestMode,
            $domaineException,
            false
        );
    }

    /**
     * Retourne un HTML autonome dont tous les éléments (images, CSS, JS...) sont encodés en base64.
     * Le HTML ainsi retourné ne dépend pas de fichiers copiés dans un dossier local ou distant.
     *
     * @param string $htmlIn
     * @param string $projectRootPath Chemin racine (pour retrouver les fichiers locaux)
     * @param array  $pdfOptions      Liste des options de configuration dédiées au module et à la conversion html
     * @param bool   $isTestMode      Indique si on est en mode test (utilisé si besoin pour votre logique)
     * @param array  $domaineException Liste des domaines à ne pas traiter
     *
     * @return string
     * @throws Exception si le fichier n'existe pas ou si l'URL n'est pas valide
     */
    public function getCleanHtmlBase64(
        string $htmlIn,
        string $projectRootPath,
        array $pdfOptions = [],
        bool $isTestMode = false,
        array $domaineException = []
    ): string {
        // On n’a pas besoin de $webPath ni de $webPathUrl
        return $this->doCleanHtml(
            $htmlIn,
            $projectRootPath,
            '',          // pas de copie de fichiers => pas de webPath
            '',          // pas de webPathUrl
            $pdfOptions,
            $isTestMode,
            $domaineException,
            true         // $useBase64 = true
        );
    }

    /**
     * Méthode commune de traitement HTML. Elle factorise un maximum de logique (parsing, transformations…).
     *
     * @param string $htmlIn
     * @param string $projectRootPath
     * @param string $webPath
     * @param string $webPathUrl
     * @param array  $pdfOptions
     * @param bool   $isTestMode
     * @param array  $domaineException
     * @param bool   $useBase64   Si true, ressources encodées en base64 (aucune copie de fichier). Sinon, copie dans $webPath.
     *
     * @return string
     * @throws Exception
     */
    private function doCleanHtml(
        string $htmlIn,
        string $projectRootPath,
        string $webPath,
        string $webPathUrl,
        array $pdfOptions,
        bool $isTestMode,
        array $domaineException,
        bool $useBase64
    ): string {
        // ---------- 1) Contrôles et normalisation des paramètres ----------
        if (!$htmlIn) {
            throw new Exception("Le contenu HTML est vide.");
        }
        if (!$projectRootPath) {
            throw new Exception("Le chemin racine du projet n'est pas renseigné.");
        }
        if (!$useBase64) {
            // Dans la version "historique", on vérifie / normalise encore $webPath et $webPathUrl
            if (!$webPath) {
                throw new Exception("Le chemin vers le répertoire web n'est pas renseigné.");
            }
            if (!$webPathUrl) {
                throw new Exception("L'url vers le répertoire web n'est pas renseigné.");
            }
            // Vérifie la validité de $webPathUrl
            if (!$this->isValidUrl($webPathUrl)) {
                throw new Exception("L'url n'est pas valide pour : " . $webPathUrl);
            }
        }

        if (!is_array($domaineException)) {
            $domaineException = [];
        }
        // On uniformise les domaines
        $domaineException = array_map(fn($domain) => strtolower(trim($domain)), $domaineException);

        // Ajout du slash de fin sur ces chemins
        if (substr($projectRootPath, -1) !== '/') {
            $projectRootPath .= '/';
        }
        if (!$useBase64) {
            // Historique => copie dans $webPath
            if (substr($webPath, -1) !== '/') {
                $webPath .= '/';
            }
            if (substr($webPathUrl, -1) !== '/') {
                $webPathUrl .= '/';
            }
            // Recréation d'un répertoire temporaire
            if (is_dir($webPath)) {
                $this->deleteDir($webPath);
            }
            if (!mkdir($webPath, 0777, true)) {
                throw new Exception("Le répertoire web n'a pas pu être créé.");
            }
        }

        // ---------- 2) Chargement du HTML dans DOMDocument ----------
        libxml_use_internal_errors(true);
        $dom = new \DOMDocument();
        $dom->loadHTML($htmlIn);

        // ---------- 3) Gestion des balises à traiter (img/src, script/src, link/href...) ----------
        // => Si ce tableau change, il est commun aux deux méthodes (maintenabilité).
        $tagAttributes = [
            'img'         => 'src',
            'script'      => 'src',
            'link'        => 'href',
            'page_header' => '',
            'page_footer' => '',
            'page'        => '',
        ];

        foreach ($tagAttributes as $tag => $attribute) {
            $elements = $dom->getElementsByTagName($tag);
            if ($elements->length === 0) {
                continue;
            }
            foreach ($elements as $element) {
                // --- 3a) Conversion des balises <page>, <page_header>, <page_footer> en <div> ---
                if (in_array($tag, ['page', 'page_header', 'page_footer'])) {
                    $this->transformCustomTagToDiv($dom, $element, $tag);
                    continue;
                }
                // --- 3b) Traitement des balises classiques (img, script, link) ---
                if (!$attribute) {
                    continue;
                }
                if (!$element instanceof \DOMElement) {
                    continue;
                }
                $filePath = $element->getAttribute($attribute);
                if (empty($filePath)) {
                    continue;
                }
                // Cas URL externe ?
                if ($this->isValidUrl($filePath)) {
                    if ($useBase64) {
                        // On remplace l'attribut par du base64 (sauf si domaine dans $domaineException)
                        $this->handleExternalResourceAsBase64($element, $attribute, $filePath, $domaineException, $isTestMode);
                    } else {
                        // Historique => copier la ressource dans $webPath
                        $this->handleExternalResource(
                            $element,
                            $attribute,
                            $filePath,
                            $domaineException,
                            $webPath,
                            $isTestMode
                        );
                    }
                } else {
                    // Cas local
                    if ($useBase64) {
                        $this->handleLocalResourceAsBase64($element, $attribute, $filePath, $projectRootPath);
                    } else {
                        $this->handleLocalResource(
                            $element,
                            $attribute,
                            $filePath,
                            $projectRootPath,
                            $webPath,
                            $webPathUrl,
                            $isTestMode
                        );
                    }
                }
            }
        }

        // ---------- 4) Sauvegarde du DOM transformé en HTML pour traiter les background-images ----------
        $htmlContent = $dom->saveHTML();

        // ---------- 5) Gestion des background-image: url() dans le style ----------
        $backgroundUrls = $this->extractAndCheckBackgroundUrls($htmlContent);
        foreach ($backgroundUrls as $backgroundImage) {
            if ($this->isValidUrl($backgroundImage)) {
                if ($useBase64) {
                    // Remplacer par base64 (sauf domaine exception)
                    $htmlContent = $this->handleBackgroundExternalAsBase64($backgroundImage, $htmlContent, $domaineException);
                } else {
                    $htmlContent = $this->handleBackgroundExternal($backgroundImage, $htmlContent, $domaineException, $webPath);
                }
            } else {
                // Cas local
                if ($useBase64) {
                    $htmlContent = $this->handleBackgroundLocalAsBase64($backgroundImage, $htmlContent, $projectRootPath);
                } else {
                    $htmlContent = $this->handleBackgroundLocal($backgroundImage, $htmlContent, $projectRootPath, $webPath);
                }
            }
        }

        // ---------- 6) Injection des styles éventuels selon les pdfOptions ----------
        $htmlContent = $this->injectStyleOptions($htmlContent, $pdfOptions);

        // ---------- 7) Vérification des éventuelles erreurs de parsing ----------
        $libxmlErrors = libxml_get_errors();
        foreach ($libxmlErrors as $error) {
            if ($error->level === LIBXML_ERR_FATAL) {
                throw new Exception(
                    "Erreur survenue lors de la manipulation du HTML : " .
                        $error->level . " -> " . $error->message
                );
            }
        }
        libxml_clear_errors();
        libxml_use_internal_errors(false);

        return $htmlContent;
    }

    /* ---------------------------------------------------------------------------
 * Les méthodes suivantes illustrent une approche possible pour gérer le flux
 * base64. Elles peuvent être ajustées selon votre code existant.
 * --------------------------------------------------------------------------- */

    /**
     * Lit un fichier local en flux et le convertit en base64 data URI.
     * @param string $absolutePath
     * @return string
     * @throws Exception
     */
    private function streamFileToBase64(string $absolutePath): string
    {
        if (!file_exists($absolutePath)) {
            throw new \Exception("Fichier introuvable : " . $absolutePath);
        }
        $mimeType = mime_content_type($absolutePath) ?: 'application/octet-stream';

        // Lecture en flux pour éviter de tout charger en mémoire
        $handle = fopen($absolutePath, 'rb');
        if (!$handle) {
            throw new \Exception("Impossible d'ouvrir le fichier : " . $absolutePath);
        }

        // Astuce : on encode par morceaux pour limiter la mémoire. 
        // Mais il faut bufferiser l’encodage base64 (pour recréer un flux final).
        $base64Data = '';
        while (!feof($handle)) {
            $chunk = fread($handle, 8192);
            // On assemble le base64 du chunk
            $base64Data .= base64_encode($chunk);
        }
        fclose($handle);

        return "data:{$mimeType};base64,{$base64Data}";
    }

    /**
     * Ex pour gérer un resource local -> attribut en base64
     */
    private function handleLocalResourceAsBase64(\DOMElement $element, string $attribute, string $filePath, string $projectRootPath): void
    {
        $absolutePath = $projectRootPath . ltrim($filePath, '/');
        $base64Uri    = $this->streamFileToBase64($absolutePath);
        $element->setAttribute($attribute, $base64Uri);
    }

    /**
     * Ex pour gérer un resource externe -> attribut en base64 (sauf si domaine exclu)
     */
    private function handleExternalResourceAsBase64(
        \DOMElement $element,
        string $attribute,
        string $url,
        array $domaineException,
        bool $isTestMode
    ): void {
        // Si le domaine est dans les exceptions ou si on est en test => on laisse l’URL telle quelle
        if (!$this->isDomainAllowed($url, $domaineException) || $isTestMode) {
            return;
        }
        // Sinon, on télécharge en streaming (curl, file_get_contents en flux, etc.)
        // Ici, pour l’exemple, on fait simple : on utilise @file_get_contents
        $raw = @file_get_contents($url);
        if (false === $raw) {
            // Vous pouvez gérer l'erreur comme vous le souhaitez
            return;
        }
        $finfo   = new \finfo(FILEINFO_MIME_TYPE);
        $mimeType = $finfo->buffer($raw) ?: 'application/octet-stream';
        $base64   = base64_encode($raw);
        $element->setAttribute($attribute, "data:{$mimeType};base64,{$base64}");
    }

    /**
     * Gestion background local -> base64
     */
    private function handleBackgroundLocalAsBase64(string $backgroundImage, string $htmlContent, string $projectRootPath): string
    {
        $absolutePath = $projectRootPath . ltrim($backgroundImage, '/');
        try {
            $base64Uri = $this->streamFileToBase64($absolutePath);
            return str_replace("url({$backgroundImage})", "url('{$base64Uri}')", $htmlContent);
        } catch (\Exception $e) {
            // Vous pouvez gérer l'erreur (ex: laisser la valeur initiale)
            return $htmlContent;
        }
    }

    /**
     * Gestion background externe -> base64
     */
    private function handleBackgroundExternalAsBase64(
        string $backgroundImage,
        string $htmlContent,
        array $domaineException
    ): string {
        // Si le domaine est exclu, on ne touche pas
        if (!$this->isDomainAllowed($backgroundImage, $domaineException)) {
            return $htmlContent;
        }
        // Sinon, on tente de récupérer la ressource
        $raw = @file_get_contents($backgroundImage);
        if (false === $raw) {
            return $htmlContent;
        }
        $finfo   = new \finfo(FILEINFO_MIME_TYPE);
        $mimeType = $finfo->buffer($raw) ?: 'application/octet-stream';
        $base64   = base64_encode($raw);
        $base64Uri = "data:{$mimeType};base64,{$base64}";

        return str_replace("url({$backgroundImage})", "url('{$base64Uri}')", $htmlContent);
    }

    /**
     * Transforme les balises <page>, <page_header>, <page_footer> en <div>.
     */
    private function transformCustomTagToDiv(\DOMDocument $dom, \DOMNode $element, string $tagName): void
    {
        $newElement = $dom->createElement('div');
        $newElement->setAttribute('id', $tagName);
        $newElement->setAttribute('class', $tagName);

        // Récupération du style
        $styleAttr = $element->attributes->getNamedItem('style');
        if ($styleAttr) {
            $newElement->setAttribute('style', $styleAttr->nodeValue);
        }

        // Convertit les attributs backtop, backbottom, backleft, backright en style
        foreach ($element->attributes as $attrName => $attrNode) {
            if (in_array($attrName, ['backtop', 'backbottom', 'backleft', 'backright'])) {
                $marginCss = match ($attrName) {
                    'backtop'    => 'margin-top',
                    'backbottom' => 'margin-bottom',
                    'backleft'   => 'margin-left',
                    'backright'  => 'margin-right',
                };
                $updatedStyle = $newElement->getAttribute('style') . $marginCss . ':' . $attrNode->nodeValue . ';';
                $newElement->setAttribute('style', $updatedStyle);
            }
        }

        // On déplace les enfants
        while ($element->firstChild) {
            $newElement->appendChild($element->firstChild);
        }
        // On remplace la balise dans le DOM
        $element->parentNode->replaceChild($newElement, $element);
    }

    /**
     * Gère le cas d'une ressource externe (URL).
     */
    private function handleExternalResource(
        \DOMElement $element,
        string      $attribute,
        string      $url,
        array       $domaineException,
        string      $webPath,
        bool        $isTestMode
    ): void {
        // Domaine autorisé ?
        if ($this->isDomainAllowed($url, $domaineException)) {
            // Vérifie que l'image existe
            if (!$this->urlExists($url)) {
                // => message rétabli comme l’attend le test unitaire
                throw new Exception("Impossible d'accéder à cette image : " . $url);
            }
            // On garde l'URL telle quelle
            $element->setAttribute($attribute, $url);
        } else {
            // Domaine non autorisé : on télécharge
            $this->downloadAndSetLocal($element, $attribute, $url, $webPath, $isTestMode);
        }
    }

    /**
     * Gère le cas d'une ressource locale (chemin relatif ou absolu).
     */
    private function handleLocalResource(
        \DOMElement $element,
        string      $attribute,
        string      $filePath,
        string      $projectRootPath,
        string      $webPath,
        string      $webPathUrl,
        bool        $isTestMode
    ): void {
        // Nettoyage du path
        $filePath = $this->cleanPath($filePath);
        $filePath = $this->ensureFullLocalPath($filePath, $projectRootPath);

        if (!file_exists($filePath)) {
            throw new Exception("Le fichier n'existe pas : " . $filePath);
        }

        // Nouveau nom
        $newBasename = 'url_to_pdf_image_' . self::hashUrl($filePath) . '.' . pathinfo($filePath, PATHINFO_EXTENSION);
        if ($isTestMode) {
            // En test, on conserve le nom d'origine
            $newBasename = pathinfo($filePath, PATHINFO_BASENAME);
        }

        $newFilePath = $webPath . $newBasename;
        if (!copy($filePath, $newFilePath)) {
            throw new Exception("Le fichier n'a pas pu être copié : " . $filePath);
        }
        chmod($newFilePath, 0777);

        // Pour certains tests, l'URL finale est aussi validée via regex (dans le code d’origine)
        // => on reproduit la vérification si nécessaire
        $formatPath = "/^(https?|ftp|s3):\/\/(localhost|[\da-z\.-]+)(\.[a-z\.]{2,6})?(:\d{2,5})?([\/\w \.-]*)*\/?$/";
        $newRelativePath = $webPathUrl . $newBasename;
        if (!preg_match($formatPath, $newRelativePath)) {
            throw new Exception("L'url n'est pas valide pour : " . $newRelativePath);
        }

        // On met à jour la balise
        $element->setAttribute($attribute, $newBasename);
    }

    /**
     * Télécharge la ressource depuis l'URL, détermine l'extension et met à jour l'élément DOM.
     */
    private function downloadAndSetLocal(\DOMElement $element, string $attribute, string $fileUrl, string $webPath, bool $isTestMode): void
    {
        $fileExist       = $this->getImageInfo($fileUrl);
        $nameFileFromUrl = $this->getFileName($fileUrl);
        $extension       = $this->getExtensionFromContentType($fileExist['content_type'], $nameFileFromUrl);

        if (!$extension) {
            throw new Exception("Type de fichier non supporté : {$fileUrl} ({$fileExist['content_type']})");
        }

        $newBasename = 'url_to_pdf_image_' . self::hashUrl($fileUrl) . '.' . $extension;
        $this->downloadFile($fileUrl, $webPath . $newBasename);
        chmod($webPath . $newBasename, 0777);

        // => message rétabli comme l’attend le test unitaire
        //    (si on fait la vérif après téléchargement, c'est surtout pour montrer la symétrie du code,
        //     mais concrètement handleExternalResource a déjà fait un urlExists.)
        if (!file_exists($webPath . $newBasename)) {
            throw new Exception("Impossible d'accéder à cette image : " . $fileUrl);
        }

        $element->setAttribute($attribute, $newBasename);
    }

    /**
     * Gère le remplacement d'une background-image externe.
     */
    private function handleBackgroundExternal(
        string $backgroundImage,
        string $htmlContent,
        array  $domaineException,
        string $webPath
    ): string {
        if ($this->isDomainAllowed($backgroundImage, $domaineException)) {
            if (!$this->urlExists($backgroundImage)) {
                throw new Exception("Impossible d'accéder à cette image : " . $backgroundImage);
            }
            // Domaine autorisé : on laisse l’URL telle quelle
            return $htmlContent;
        }

        // Domaine non autorisé : on télécharge et on remplace
        $fileExist  = $this->getImageInfo($backgroundImage);
        $nameFile   = $this->getFileName($backgroundImage);
        $extension  = $this->getExtensionFromContentType($fileExist['content_type'], $nameFile);
        if (!$extension) {
            throw new Exception("Type de fichier non supporté pour : {$backgroundImage}");
        }

        $newBasename = 'url_to_pdf_image_' . self::hashUrl($backgroundImage) . '.' . $extension;
        $this->downloadFile($backgroundImage, $webPath . $newBasename);
        chmod($webPath . $newBasename, 0777);

        return str_replace($backgroundImage, $newBasename, $htmlContent);
    }

    /**
     * Gère le remplacement d'une background-image locale.
     */
    private function handleBackgroundLocal(
        string $backgroundImage,
        string $htmlContent,
        string $projectRootPath,
        string $webPath
    ): string {
        $localPath = $this->cleanPath($backgroundImage);
        $localPath = $this->ensureFullLocalPath($localPath, $projectRootPath);

        if (!file_exists($localPath)) {
            throw new Exception("Le fichier n'existe pas : " . $localPath);
        }

        $newBasename = 'url_to_pdf_image_' . self::hashUrl($localPath) . '.' . pathinfo($localPath, PATHINFO_EXTENSION);
        if (!copy($localPath, $webPath . $newBasename)) {
            throw new Exception("Le fichier n'a pas pu être copié : " . $localPath);
        }
        chmod($webPath . $newBasename, 0777);

        return str_replace($backgroundImage, $newBasename, $htmlContent);
    }

    /**
     * Nettoie le chemin relatif en supprimant '../', './' et '//' redondants.
     */
    private function cleanPath(string $path): string
    {
        $path = str_replace(['../', './'], '', $path);
        // Remplace les doublons de slash par un seul slash
        return str_replace('//', '/', $path);
    }

    /**
     * Force le chemin local à être complet en le préfixant du $projectRootPath si nécessaire.
     */
    private function ensureFullLocalPath(string $filePath, string $projectRootPath): string
    {
        // Si on commence par un / et que le chemin ne commence pas déjà par $projectRootPath
        if (str_starts_with($filePath, '/') && !str_starts_with($filePath, $projectRootPath)) {
            $filePath = ltrim($filePath, '/');
        }

        // On récupère le chemin réel si possible
        if (realpath($filePath)) {
            $filePath = realpath($filePath);
        }

        // Si le fichier n'est pas déjà dans $projectRootPath, on le force
        if (!str_starts_with($filePath, $projectRootPath)) {
            $filePath = rtrim($projectRootPath, '/') . '/' . ltrim($filePath, '/');
        }

        return $filePath;
    }

    /**
     * Injecte (ou complète) un bloc <style> pour gérer certaines options (font, noBreakTableRows…).
     */
    private function injectStyleOptions(string $htmlContent, array $pdfOptions): string
    {
        $styleAttributes = '';

        if (!empty($pdfOptions['noBreakTableRows'])) {
            $styleAttributes .= 'tr, td, th { page-break-inside: avoid; }';
        }
        if (!empty($pdfOptions['font'])) {
            $styleAttributes .= 'html { ';
            foreach ($pdfOptions['font'] as $font => $att) {
                $styleAttributes .= 'font-' . $font . ': ' . $att . '; ';
            }
            $styleAttributes .= '}';
        }

        // Pas de style à injecter ?
        if (empty($styleAttributes)) {
            return $htmlContent;
        }

        // On ré-ouvre le DOM pour injecter le style
        $dom = new \DOMDocument();
        $dom->loadHTML($htmlContent);

        $styleElement = null;
        if ($dom->getElementsByTagName('style')->length === 0) {
            // On crée le noeud style
            $styleElement = $dom->createElement('style', $styleAttributes);
        } else {
            // On complète le premier <style> trouvé
            $styleElement = $dom->getElementsByTagName('style')->item(0);
            $newStyleElement = $dom->createElement('style', $styleElement->nodeValue . $styleAttributes);
            $styleElement->parentNode->replaceChild($newStyleElement, $styleElement);
        }

        // S'il existe un <head>, on y ajoute le <style>
        if ($dom->getElementsByTagName('head')->length > 0) {
            $dom->getElementsByTagName('head')->item(0)->appendChild($styleElement);
        } else {
            // Sinon on crée un <head> et on y glisse <style>
            $head = $dom->createElement('head');
            $htmlTag = $dom->getElementsByTagName('html')->item(0);
            if ($htmlTag && $htmlTag->hasChildNodes()) {
                $firstChild = $htmlTag->firstChild;
                $htmlTag->insertBefore($head, $firstChild);
            } else {
                $htmlTag->appendChild($head);
            }
            $head->appendChild($styleElement);
        }

        return $dom->saveHTML();
    }


    /**
     * Supprime les images temporaires
     * 
     * @param string $webPath Chemin vers le répertoire web (ex: /var/www/html/public-data/tmp/url_to_pdf/)
     * @return void
     */
    public function deleteTempImages(string $webPath): void
    {
        if (substr($webPath, -1) !== '/') {
            $webPath .= '/';
        }

        if (is_dir($webPath)) {
            $this->rrmdir($webPath);
        }
    }

    private function rrmdir(string $dir): void
    {
        if (is_dir($dir)) {
            $objects = scandir($dir);
            foreach ($objects as $object) {
                if ($object != "." && $object != "..") {
                    if (is_dir($dir . "/" . $object)) {
                        $this->rrmdir($dir . "/" . $object);
                    } else {
                        unlink($dir . "/" . $object);
                    }
                }
            }
            rmdir($dir);
        }
    }

    /**
     * Supprime un dossier et son contenu m^me si non vide
     * 
     * @param string $dir Chemin vers le répertoire à supprimer
     * @return bool (true si le dossier a été supprimé)
     */
    public function deleteDir(string $dir): bool
    {
        if (!is_dir($dir)) {
            return false;
        }

        $files = array_diff(scandir($dir), ['.', '..']);
        foreach ($files as $file) {
            (is_dir("$dir/$file")) ? $this->deleteDir("$dir/$file") : unlink("$dir/$file");
        }

        if (!rmdir($dir)) {
            throw new Exception("Le répertoire " . $dir . " n'a pas pu être supprimé.");
        }
        return true;
    }
    /**
     * Retourne un PDF à partir de HTML "inline" (ressources en base64 ou depuis le Web).
     *
     * @param string $htmlContent    Contenu HTML complet à convertir
     * @param array  $loadingOptions Options à appliquer pour le chargement de la page (ex: waitUntil, timeout, etc.)
     * @param array  $pdfOptions     Options pour la génération du PDF (ex: format, noBreakTableRows, etc.)
     * @param array  $authOptions    Identifiants basiques à transmettre (ex: ['username'=>'...', 'password'=>'...'])
     * @return resource Stream contenant le PDF
     * @throws \Exception si cURL échoue, si la réponse n'est pas un PDF ou si $htmlContent est vide
     */

    public function getPdfFromHtmlContent(
        string $htmlContent,
        array  $loadingOptions = [],
        array  $pdfOptions = [],
        array  $authOptions = []
    ) {
        // Vérification minimale du HTML
        if (!$htmlContent) {
            throw new Exception("Le contenu HTML est vide.");
        }

        // Préparation de la charge utile pour l'API
        $payload = [
            'html'   => $htmlContent,
            'loadingOptions' => $loadingOptions ?: ['waitUntil' => 'networkidle2'],
            'pdfOptions'    => $pdfOptions     ?: ['format' => 'a4'],
            'auth'          => $authOptions,
            'urlMustHave200' => $pdfOptions['urlMustHave200'] ?? false,
        ];

        // Création du stream de sortie
        $stream = fopen('php://temp', 'r+');
        if ($stream === false) {
            throw new Exception("Impossible de créer le stream temporaire.");
        }

        // Préparation de la requête cURL
        $ch = curl_init();
        $headers = [
            'Content-Type: application/json',
            'ACCESS-KEY: ' . $this->accessKey,
        ];
        curl_setopt($ch, CURLOPT_URL, $this->apiUrl . '/get-pdf');
        curl_setopt($ch, CURLOPT_POST, true);
        curl_setopt($ch, CURLOPT_POSTFIELDS, json_encode($payload));
        curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
        curl_setopt($ch, CURLOPT_FILE, $stream);

        // Exécution
        $success = curl_exec($ch);
        if ($success === false) {
            $error = curl_error($ch);
            curl_close($ch);
            fclose($stream);
            throw new Exception("Erreur cURL : " . $error);
        }

        $httpCode    = curl_getinfo($ch, CURLINFO_HTTP_CODE);
        $contentType = curl_getinfo($ch, CURLINFO_CONTENT_TYPE);
        curl_close($ch);

        // Vérifications de base
        if ($httpCode !== 200) {
            fclose($stream);
            throw new Exception("Erreur HTTP $httpCode");
        }

        // Vérifie si le contenu retourné est effectivement un PDF
        rewind($stream);
        $pdfHeader = fread($stream, 4);
        if (empty($pdfHeader) || $pdfHeader !== '%PDF') {
            fclose($stream);
            throw new Exception("La réponse retournée est vide ou n'est pas un PDF.");
        }

        // Remet le pointeur au début du stream
        rewind($stream);
        return $stream;
    }

    /**
     * Affiche un flux PDF directement dans le navigateur avec les en-têtes appropriés.
     *
     * @param resource $pdfStream Le flux du document PDF à afficher. Doit être une ressource valide.
     * @param string $filename (optionnel) Le nom du fichier à afficher dans le navigateur. Par défaut : 'document.pdf'.
     * @param int $cacheDuration (optionnel) La durée de vie du cache en secondes. Par défaut : 15 minutes (900 secondes).
     *
     * @throws Exception Si le paramètre $pdfStream n'est pas une ressource valide.
     * @return void Cette méthode n'a pas de retour, elle termine l'exécution du script après avoir affiché le PDF.
     */
    public static function showPdfStream($pdfStream, string $filename = 'document.pdf', int $cacheDuration = 900): void
    {
        if (!is_resource($pdfStream)) {
            throw new Exception('showPdfStream(): The provided argument is not a valid resource.');
        }

        header('Content-Type: application/pdf');
        header('Content-Disposition: inline; filename="' . $filename . '"');
        header('Cache-Control: public, max-age=' . $cacheDuration);
        header('Pragma: public');
        header('Expires: ' . gmdate('D, d M Y H:i:s', time() + $cacheDuration) . ' GMT');
        fpassthru($pdfStream);
        fclose($pdfStream);
        exit;
    }
}
